# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Link store and field for generating links."""

from collections import defaultdict

from marshmallow import Schema, fields, missing, post_dump
from uritemplate import URITemplate


class LinksStore:
    """Utility class for keeping track of links.

    The ``config`` argument is a dictionary with link namespaces as keys and a
    dictionary of link types and their URITemplate objects.

    Exmple usage:

    ..code-block:: python

        link_store = LinksStore()
        data = {
            'links': {
                'self': {'pid_value': '12345'},
                'draft': {'pid_value': '12345'},
            },
        }
        link_store.add('record', data['links'])
        link_store.resolve(config={
            'record': {
                'self': URITemplate('/api/records{/pid_value}'),
                'draft': URITemplate('/api/records{/pid_value}/draft'),
            }
        })
    """

    def __init__(self, host=None, config=None):
        """Constructor."""
        self._host = host
        self.config = config
        self._links = defaultdict(list)

    def add(self, namespace, links):
        """Adds a dictionary of links under a namespace."""
        self._links[namespace].append(links)

    @property
    def host(self):
        """Get the hostname."""
        return self._host() if callable(self._host) else self._host

    def resolve(self, context=None, config=None):
        """Resolves in-place all the tracked link dictionaries."""
        config = config or self.config
        assert config, "Links config is empty."

        context = context or {}
        parameters = self._links

        for ns, keyed_parameters_array in parameters.items():
            if ns not in config:
                raise Exception(f'Unknown links namespace: {ns}')

            # Why is this a list?
            for keyed_parameters in keyed_parameters_array:
                # Always just a list of 1 dict with
                # keys: each link key
                # values: dict with 1
                #     key: "params"
                #     value: dict of querystring parameters
                for k, params in keyed_parameters.items():
                    keyed_parameters[k] = self.to_link(
                        config[ns][k],
                        {**context, **params}
                    )

    def to_link(self, template, values):
        """Expand the link template with the values.

        NOTE: querystring parameters (``params``) are converted to associative
              arrays in order to be expanded correctly.
        """
        qs_dict = values.get("params", {})
        qs_associative_array = []
        for param, value in sorted(qs_dict.items()):
            if isinstance(value, list):
                param_array = [(param, e) for e in value]
            else:
                param_array = [(param, value)]

            qs_associative_array.extend(param_array)

        path = template.expand({**values, "params": qs_associative_array})

        return self.base_url(host=self.host, path=path)

    @staticmethod
    def base_url(scheme="https", host=None, path="/", querystring="",
                 api=False):
        """Creates the URL for API and UI endpoints."""
        assert path.startswith("/")
        path = f"/api{path}" if api else path
        return f"{scheme}://{host}{path}{querystring}"


class LinksSchema(Schema):
    """Links schema that stores the result in a link store."""

    namespace = None

    @post_dump()
    def _store(self, data, **kwargs):
        self.context['links_store'].add(self.namespace, data)
        return data



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2020 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Sanitized Unicode string field."""

from ftfy import fix_text
from marshmallow import fields

from .trimmed import TrimmedString


class SanitizedUnicode(TrimmedString):
    """String field that sanitizes and fixes problematic unicode characters."""

    UNWANTED_CHARACTERS = {
        # Zero-width space
        u'\u200b',
    }

    def is_valid_xml_char(self, char):
        """Check if a character is valid based on the XML specification."""
        codepoint = ord(char)
        return (0x20 <= codepoint <= 0xD7FF or
                codepoint in (0x9, 0xA, 0xD) or
                0xE000 <= codepoint <= 0xFFFD or
                0x10000 <= codepoint <= 0x10FFFF)

    def _deserialize(self, value, attr, data, **kwargs):
        """Deserialize sanitized string value."""
        value = super()._deserialize(
            value, attr, data, **kwargs)
        value = fix_text(value)

        # NOTE: This `join` might be ineffiecient... There's a solution with a
        # large compiled regex lying around, but needs a lot of tweaking.
        value = ''.join(filter(self.is_valid_xml_char, value))
        for char in self.UNWANTED_CHARACTERS:
            value = value.replace(char, '')
        return value



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Test LinksStore."""


from uritemplate import URITemplate

from marshmallow_utils.links import LinksStore


def test_resolve_params():
    links = LinksStore(host="localhost")
    links_config = {
        "search": {
            "self": URITemplate("/{?params*}"),
        }
    }
    data = {
        "self": {
            "params": {
                "type": ["A", "B"],
                "sort": "newest",
                "subtype": ["1"],
                "size": 10,
            }
        }
    }
    links.add("search", data)

    links.resolve(config=links_config)

    assert (
        "https://localhost/?size=10&sort=newest&subtype=1&type=A&type=B" ==
        data["self"]
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2020 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Extended Date(/Time) Format Level 0 date string field."""

from datetime import date

from edtf.parser.grammar import ParseException, level0Expression
from marshmallow import fields


def _(x):
    return x


class EDTFDateString(fields.Str):
    """
    Extended Date(/Time) Format Level 0 date string field.

    Made a field for stronger semantics than just a validator.
    """

    default_error_messages = {
        "invalid": _("Please provide a valid date or interval.")
    }

    def _deserialize(self, value, attr, data, **kwargs):
        """Deserialize an EDTF Level 0 formatted date string.

        load()-equivalent operation.

        NOTE: Level 0 allows for an interval.
        NOTE: ``level0Expression`` tries hard to parse dates. For example,
              ``"2020-01-02garbage"`` will parse to the 2020-01-02 date.
        """
        parser = level0Expression("level0")

        try:
            result = parser.parseString(value)

            if not result:
                raise ParseException()

            # check it is chronological if interval
            # NOTE: EDTF Date and Interval both have same interface
            #       and date.lower_strict() <= date.upper_strict() is always
            #       True for a Date
            result = result[0]
            if result.upper_strict() < result.lower_strict():
                raise self.make_error("invalid")

            return (
                super(EDTFDateString, self)
                ._deserialize(str(result), attr, data, **kwargs)
            )
        except ParseException:
            raise self.make_error("invalid")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Link store and field for generating links."""

from marshmallow import fields, missing


class LinksField(fields.Field):
    """Links field."""

    # NOTE: forces serialization
    _CHECK_ATTRIBUTE = False

    def __init__(self, links_schema=None, namespace=None, **kwargs):
        """Constructor."""
        self.links_schema = links_schema
        self.namespace = namespace
        super().__init__(**kwargs)

    def _serialize(self, value, attr, obj, *args, **kwargs):
        # NOTE: We pass the full parent `obj`, since we want to make it
        # available to the links schema
        result = self.links_schema(context=self.context).dump(obj)
        # NOTE: By adding the "result" dictionary to the link store we achieve
        # that the link store holds a reference to this dictionary which we
        # also return as the result for this field. This is critically in
        # making the LinksStore.resolve() work. Because the link store holds a
        # reference, it doesn't matter where the links are injected in the
        # final record projection, because any updates made to the result
        # in the link store will automatically be available in the record
        # projection because it's the same dictionary being modified.
        self.context['links_store'].add(self.namespace, result)
        return result

    def _deserialize(self, value, attr, data, **kwargs):
        # NOTE: we don't deserialize the links, and we don't use dump_only=True
        # because that will by default raise a validation error unless you have
        # configured marshmallow to exclude unknown values.
        return missing



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2020 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Marshmallow fields."""

from .edtfdatestring import EDTFDateString
from .generated import GenFunction, GenMethod
from .isodate import ISODateString
from .isolanguage import ISOLangString
from .links import LinksField
from .nestedattr import NestedAttribute
from .sanitizedhtml import ALLOWED_HTML_ATTRS, ALLOWED_HTML_TAGS, SanitizedHTML
from .sanitizedunicode import SanitizedUnicode
from .trimmed import TrimmedString

__all__ = (
    'ALLOWED_HTML_ATTRS',
    'ALLOWED_HTML_TAGS',
    'EDTFDateString',
    'GenFunction',
    'GenMethod',
    'ISODateString',
    'ISOLangString',
    'LinksField',
    'NestedAttribute',
    'SanitizedHTML',
    'SanitizedUnicode',
    'TrimmedString',
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2020 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""HTML sanitized string field."""

import bleach

from .sanitizedunicode import SanitizedUnicode

#: Allowed tags used for html sanitizing by bleach.
ALLOWED_HTML_TAGS = [
    'a',
    'abbr',
    'acronym',
    'b',
    'blockquote',
    'br',
    'code',
    'div',
    'em',
    'h1',
    'h2',
    'h3',
    'h4',
    'h5',
    'i',
    'li',
    'ol',
    'p',
    'pre',
    'span',
    'strike',
    'strong',
    'sub',
    'sup',
    'u',
    'ul',
]

#: Allowed attributes used for html sanitizing by bleach.
ALLOWED_HTML_ATTRS = {
    '*': ['class'],
    'a': ['href', 'title', 'name', 'class', 'rel'],
    'abbr': ['title'],
    'acronym': ['title'],
}


class SanitizedHTML(SanitizedUnicode):
    """String field which sanitizes HTML using the bleach library.

    The default list of allowed tags and attributes is defined by
    ``ALLOWED_HTML_TAGS`` and ``ALLOWED_HTML_ATTRS``.

    You can override the defaults like this:

    .. code-block:: python

        class MySchema(Schema):
            html = fields.SanitizedHTML(tags=['a'], attrs={'a': ['href']})

    :param tags: List of allowed tags.
    :param attrs: Dictionary of allowed attributes per tag.
    """

    def __init__(self, tags=None, attrs=None, *args, **kwargs):
        """Initialize field."""
        super().__init__(*args, **kwargs)
        self.tags = tags or ALLOWED_HTML_TAGS
        self.attrs = attrs or ALLOWED_HTML_ATTRS

    def _deserialize(self, value, attr, data, **kwargs):
        """Deserialize string by sanitizing HTML."""
        value = super()._deserialize(
            value, attr, data, **kwargs)
        return bleach.clean(
            value,
            tags=self.tags,
            attributes=self.attrs,
            strip=True,
        ).strip()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2020 CERN.
#
# Marshmallow-Utils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Date string field."""

import arrow
from arrow.parser import ParserError
from marshmallow import fields, missing


class ISODateString(fields.Date):
    """ISO8601-formatted date string."""

    def _serialize(self, value, attr, obj, **kwargs):
        """Serialize an ISO8601-formatted date."""
        try:
            return super()._serialize(
                arrow.get(value).date(), attr, obj, **kwargs)
        except ParserError:
            return missing

    def _deserialize(self, value, attr, data, **kwargs):
        """Deserialize an ISO8601-formatted date."""
        return super()._deserialize(value, attr, data, **kwargs).isoformat()
